// SPDX-License-Identifier: BSD-3-Clause

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/sendfile.h>
#include <sys/eventfd.h>
#include <libaio.h>
#include <errno.h>

#include "aws.h"
#include "utils/util.h"
#include "utils/debug.h"
#include "utils/sock_util.h"
#include "utils/w_epoll.h"

static const char HTTP_HEADER[] = "HTTP/1.0 200 OK\r\nContent-Length: %ld\r\nConnection: close\r\n\r\n";

static const char NOT_FOUND[] = "HTTP/1.0 404 Not Found\r\n\r\n";

/* server socket file descriptor */
static int listenfd;

/* epoll file descriptor */
static int epollfd;

static int aws_on_path_cb(http_parser *p, const char *buf, size_t len)
{
	char formated_buf[BUFSIZ] = { 0 };
	struct connection *conn = (struct connection *)p->data;

	strcat(formated_buf, AWS_DOCUMENT_ROOT);
	strncat(formated_buf, buf + 1, len - 1);
	strcpy(conn->request_path, formated_buf);
	conn->request_path[strlen(AWS_DOCUMENT_ROOT) + len - 1] = '\0';
	conn->have_path = 1;

	return 0;
}

int parse_header(struct connection *conn)
{
	http_parser_settings settings_on_path = {
		.on_message_begin = 0,
		.on_header_field = 0,
		.on_header_value = 0,
		.on_path = aws_on_path_cb,
		.on_url = 0,
		.on_fragment = 0,
		.on_query_string = 0,
		.on_body = 0,
		.on_headers_complete = 0,
		.on_message_complete = 0
	};

	http_parser_init(&conn->request_parser, HTTP_REQUEST);
	conn->request_parser.data = conn;
	http_parser_execute(&conn->request_parser, &settings_on_path,
						conn->recv_buffer, strlen(conn->recv_buffer));
	return 0;
}

static enum resource_type connection_get_resource_type(struct connection *conn)
{
	conn->res_type = RESOURCE_TYPE_NONE;
	if (strstr(conn->request_path, "dynamic"))
		conn->res_type = RESOURCE_TYPE_DYNAMIC;
	if (strstr(conn->request_path, "static"))
		conn->res_type = RESOURCE_TYPE_STATIC;
	return conn->res_type;
}

static void connection_prepare_send_reply_header(struct connection *conn)
{
	connection_get_resource_type(conn);
	conn->send_len = snprintf(conn->send_buffer, BUFSIZ,
							  HTTP_HEADER, conn->file_size);
	dlog(LOG_INFO, "Set header with len %zu\n", conn->send_len);
	conn->state = STATE_SENDING_HEADER;
}

static void connection_prepare_send_404(struct connection *conn)
{
	conn->state = STATE_SENDING_404;
	strcpy(conn->send_buffer, NOT_FOUND);
	conn->send_len = strlen(NOT_FOUND);
}

int connection_open_file(struct connection *conn)
{
	struct stat stat_buf;

	conn->fd = open(conn->request_path, O_RDONLY | O_NONBLOCK);
	if (conn->fd < 0) {
		dlog(LOG_ERR, "Could not open file: %s\n", conn->request_path);
	} else {
		DIE(fstat(conn->fd, &stat_buf) < 0, "File size failed");
		conn->file_size = stat_buf.st_size;
		dlog(LOG_INFO, "Opened file %s at fd %d with size %zu\n", conn->request_path, conn->fd, conn->file_size);
	}

	return conn->fd;
}

void connection_start_async_io(struct connection *conn)
{
	int rc;

	conn->send_pos = 0;
	conn->send_len = 0;
	memset(conn->send_buffer, 0, BUFSIZ);

	memset(&conn->iocb, 0, sizeof(conn->iocb));
	dlog(LOG_INFO, "Offset at: %zu\n", conn->async_read_len);
	io_prep_pread(&conn->iocb, conn->fd, conn->send_buffer,
				  BUFSIZ, conn->async_read_len);

	conn->eventfd = eventfd(0, EFD_NONBLOCK);
	DIE(conn->eventfd < 0, "eventfd failed");
	io_set_eventfd(&conn->iocb, conn->eventfd);
	conn->piocb[0] = &conn->iocb;

	rc = io_submit(conn->ctx, 1, conn->piocb);
	if (rc < 0) {
		dlog(LOG_ERR, "io_submit failed with rc = %d, for fd = %d\n", rc, conn->fd);
		DIE(1, "io_submit failed");
	}

	dlog(LOG_INFO, "Async read started successfully\n");
	conn->state = STATE_ASYNC_ONGOING;
}

void connection_complete_async_io(struct connection *conn)
{
	struct io_event event;
	int rc;

	rc = io_getevents(conn->ctx, 1, 1, &event, NULL);
	DIE(rc < 0, "io_getevents failed\n");

	if (event.res < 0) {
		dlog(LOG_ERR, "Async read of negative bytes: %zd\n", event.res);
		DIE(1, "Async read failed");
	} else if (event.res > 0) {
		dlog(LOG_INFO, "Async read of %zd bytes completed\n", event.res);
		conn->send_len += event.res;
		DIE(close(conn->eventfd) < 0, "could not close eventfd");
		conn->state = STATE_SENDING_DATA;
	}
}

void handle_new_connection(void)
{
	static int sockfd;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	struct sockaddr_in addr;
	struct connection *conn;
	int rc;

	sockfd = accept(listenfd, (SSA *) &addr, &addrlen);
	DIE(sockfd < 0, "Could not accept client");

	int flags = fcntl(sockfd, F_GETFL, 0);

	DIE(flags < 0, "Could not get flags for new sockfd");
	rc = fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);
	DIE(rc < 0, "Could not set new sockfd to nonblocking");

	conn = connection_create(sockfd);

	memset(&conn->ctx, 0, sizeof(conn->ctx));
	rc = io_setup(1, &conn->ctx);
	if (rc < 0) {
		dlog(LOG_ERR, "io_setup failed with rc = %d,for fd = %d\n", rc, conn->fd);
		DIE(rc < 0, "io_setup failed");
	}

	dlog(LOG_INFO, "Accepted connection %d from: %s:%d\n", conn->sockfd,
		 inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

	rc = w_epoll_add_ptr_inout(epollfd, sockfd, conn);
	DIE(rc < 0, "w_epoll_add_ptr_in");
}

struct connection *connection_create(int sockfd)
{
	struct connection *conn = calloc(1, sizeof(*conn));

	DIE(!conn, "Connection allocation failed");

	conn->sockfd = sockfd;
	memset(conn->recv_buffer, 0, BUFSIZ);
	memset(conn->send_buffer, 0, BUFSIZ);
	conn->state = STATE_INITIAL;

	return conn;
}

void connection_remove(struct connection *conn)
{
	DIE(w_epoll_remove_ptr(epollfd, conn->sockfd, conn) < 0,
		"Could not remove sockfd from epoll");
	DIE(tcp_close_connection(conn->sockfd) < 0, "Could not close connection");
	DIE(io_destroy(conn->ctx) != 0, "Could not destroy io ctx");
	dlog(LOG_INFO, "Client disconnected\n");
	conn->state = STATE_CONNECTION_CLOSED;
	free(conn);
}

void receive_data(struct connection *conn)
{
	int rc;
	static char buf[BUFSIZ];

	memset(buf, 0, BUFSIZ);
	rc = recv(conn->sockfd, buf, BUFSIZ, 0);
	if (rc < 0) {
		if (errno == EAGAIN || errno == EWOULDBLOCK)
			return;
		DIE(1, "recv failed");
	}

	conn->state = STATE_RECEIVING_DATA;
	if (strstr(buf, "\r\n\r\n")) {
		conn->state = STATE_REQUEST_RECEIVED;
		dlog(LOG_INFO, "Request receiving finished!\n");
	}

	conn->recv_len += rc;
	strcat(conn->recv_buffer, buf);
}

enum connection_state connection_send_static(struct connection *conn)
{
	int rc;

	rc = sendfile(conn->sockfd, conn->fd,
				  &conn->file_pos, conn->file_size - conn->file_pos);
	if (rc < 0)
		if (errno != EAGAIN)
			dlog(LOG_ERR, "sendfile failed");
	if (conn->file_pos == conn->file_size) {
		DIE(close(conn->fd), "Cannot close fd\n");
		dlog(LOG_INFO, "Finished sendfile()\n");
		conn->state = STATE_DATA_SENT;
	}
	return conn->state;
}

int connection_send_dynamic(struct connection *conn)
{
	int rc;

	rc = send(conn->sockfd, conn->send_buffer + conn->send_pos,
			  conn->send_len - conn->send_pos, 0);
	if (rc < 0) {
		if (errno != EAGAIN && errno != EWOULDBLOCK) {
			dlog(LOG_ERR, "send failed");
			return -1;
		}
	} else {
		conn->send_pos += rc;
	}
	dlog(LOG_INFO, "Sending from send_buffer (with offset %zu): %zu bytes\n",
		 conn->send_pos, conn->send_len - conn->send_pos);
	if (conn->send_pos == conn->send_len) {
		conn->async_read_len += conn->send_len;
		dlog(LOG_INFO, "Finished sending dynamic file segment\n");
		if (conn->async_read_len == conn->file_size) {
			conn->state = STATE_DATA_SENT;
			dlog(LOG_INFO, "Finished sending dynamic file fully\n");
			DIE(close(conn->fd), "Cannot close fd\n");
		} else {
			connection_start_async_io(conn);
		}
	}
	return 0;
}

void handle_input(struct connection *conn)
{
	switch (conn->state) {
	case STATE_INITIAL:
		receive_data(conn);
		break;
	case STATE_RECEIVING_DATA:
		receive_data(conn);
		break;
	default:
		break;
	}
}

void handle_output(struct connection *conn)
{
	int rc;
	uint64_t read_file;

	switch (conn->state) {
	case STATE_REQUEST_RECEIVED:
		parse_header(conn);
		rc = connection_open_file(conn);
		if (rc >= 0)
			connection_prepare_send_reply_header(conn);
		else
			connection_prepare_send_404(conn);
		break;
	case STATE_SENDING_HEADER:
		rc = send(conn->sockfd, conn->send_buffer + conn->send_pos,
				  conn->send_len - conn->send_pos, 0);
		if (rc < 0) {
			if (errno != EAGAIN && errno != EWOULDBLOCK)
				dlog(LOG_ERR, "Could not send header");
		} else {
			conn->send_pos += rc;
			if (conn->send_pos == conn->send_len) {
				switch (conn->res_type) {
				case RESOURCE_TYPE_DYNAMIC:
					dlog(LOG_INFO, "Finished sending header:\n%s",
						conn->send_buffer);
					connection_start_async_io(conn);
					break;
				case RESOURCE_TYPE_STATIC:
					dlog(LOG_INFO, "Finished sending header:\n%s",
						 conn->send_buffer);
					conn->state = STATE_SENDING_DATA;
					break;
				default:
					break;
				}
			}
		}
		break;
	case STATE_ASYNC_ONGOING:
		rc = read(conn->eventfd, &read_file, sizeof(read_file));
		DIE(rc < 0 && errno != EAGAIN, "eventfd read failed");
		if (read_file != 0)
			connection_complete_async_io(conn);
		break;
	case STATE_SENDING_DATA:
		switch (conn->res_type) {
		case RESOURCE_TYPE_DYNAMIC:
			connection_send_dynamic(conn);
			break;
		case RESOURCE_TYPE_STATIC:
			connection_send_static(conn);
			break;
		default:
			DIE(1, "Could not figure out resource type");
		}
		break;
	case STATE_SENDING_404:
		rc = send(conn->sockfd, conn->send_buffer + conn->send_pos,
				  conn->send_len - conn->send_pos, 0);
		DIE(rc < -1, "Could not send 404");
		conn->send_pos += rc;
		if (conn->send_pos == conn->send_len)
			conn->state = STATE_404_SENT;
		break;
	default:
		break;
	}
}

void handle_client(uint32_t event, struct connection *conn)
{
	if (event & EPOLLIN)
		handle_input(conn);
	if (event & EPOLLOUT)
		handle_output(conn);
	if (conn->state == STATE_DATA_SENT || conn->state == STATE_404_SENT)
		connection_remove(conn);
	DIE(event & EPOLLERR, "event error");
}

int main(void)
{
	int rc, n;

	epollfd = w_epoll_create();
	DIE(epollfd < 0, "w_epoll_create failed");

	listenfd = tcp_create_listener(AWS_LISTEN_PORT, DEFAULT_LISTEN_BACKLOG);
	DIE(listenfd < 0, "tcp_create_listener failed");

	rc = w_epoll_add_fd_in(epollfd, listenfd);
	DIE(rc < 0, "w_epoll_add_fd_in failed");

	dlog(LOG_INFO, "Server waiting for connections on port %d\n",
		 AWS_LISTEN_PORT);

	/* server main loop */
	while (1) {
		struct epoll_event rev;

		n = w_epoll_wait_infinite(epollfd, &rev);
		DIE(n < 0, "epoll_wait failed");

		if (rev.data.fd == listenfd && rev.events & EPOLLIN)
			handle_new_connection();
		else
			handle_client(rev.events, rev.data.ptr);
	}

	rc = close(listenfd);
	DIE(rc < 0, "Close listenfd");
	rc = close(epollfd);
	DIE(rc < 0, "Close epollfd");

	return 0;
}
